from juliacall import Main as jl
import numpy as np
import pandas as pd
from sklearn.model_selection import KFold

class LWPLSR:
    """The lwpls regression model from Jchemo (M. Lesnoff)

    Returns:
        self.scores (DataFrame): various metrics and scores
        self.predicted_results_on_train (DataFrame):
        self.predicted_results_on_cv (DataFrame):
        self.predicted_results_on_test (DataFrame):
        self.mod (Julia model): the prepared model
    """
    def __init__(self, x_train, y_train, x_test, y_test, x_train_cv1, y_train_cv1, x_test_cv1, y_test_cv1, x_train_cv2, y_train_cv2, x_test_cv2, y_test_cv2, x_train_cv3, y_train_cv3, x_test_cv3, y_test_cv3):
        """Initiate the LWPLSR and prepare data for Julia computing."""

        self.x_train, self.y_train, self.x_test, self.y_test = x_train, y_train, x_test, y_test
        self.x_train_cv1, self.y_train_cv1, self.x_test_cv1, self.y_test_cv1 = x_train_cv1, y_train_cv1, x_test_cv1, y_test_cv1
        self.x_train_cv2, self.y_train_cv2, self.x_test_cv2, self.y_test_cv2 = x_train_cv2, y_train_cv2, x_test_cv2, y_test_cv2
        self.x_train_cv3, self.y_train_cv3, self.x_test_cv3, self.y_test_cv3 = x_train_cv3, y_train_cv3, x_test_cv3, y_test_cv3
        # prepare to send dataframes to julia and Jchemo
        jl.x_train, jl.y_train, jl.x_test, jl.y_test = self.x_train, self.y_train, self.x_test, self.y_test
        jl.x_train_cv1, jl.y_train_cv1, jl.x_test_cv1, jl.y_test_cv1 = self.x_train_cv1, self.y_train_cv1, self.x_test_cv1, self.y_test_cv1
        jl.x_train_cv2, jl.y_train_cv2, jl.x_test_cv2, jl.y_test_cv2 = self.x_train_cv2, self.y_train_cv2, self.x_test_cv2, self.y_test_cv2
        jl.x_train_cv3, jl.y_train_cv3, jl.x_test_cv3, jl.y_test_cv3= self.x_train_cv3, self.y_train_cv3, self.x_test_cv3, self.y_test_cv3
        # optimize lwplsr parameters with Jchemo
        # jl.seval("""
        # ntrain = nro(x_train)
        # segm = segmkf(ntrain, 4; rep = 5)
        # nlvdis = [5; 10; 15] ; metric = [:mah]
        # h = [1; 2; 6; Inf] ; k = [10; 30; 100]
        # nlv = 0:15
        # pars = mpar(nlvdis = nlvdis, metric = metric, h = h, k = k)
        # println(pars)
        # """)

        # initialize vars from the class
        y_shape = y_test.shape
        y_shape_cv1 = y_test_cv1.shape
        y_shape_cv2 = y_test_cv2.shape
        y_shape_cv3 = y_test_cv3.shape
        # self.scores = pd.DataFrame
        self.predicted_results_on_test = pd.DataFrame
        self.predicted_results_on_train = pd.DataFrame
        self.predicted_results_on_cv = pd.DataFrame
        self.pred_test = np.zeros(shape=(y_shape[0], 1))
        self.pred_train = np.zeros(shape=(y_shape[0], 1))
        self.pred_cv1 = np.zeros(shape=(y_shape_cv1[0], 1))
        self.pred_cv2 = np.zeros(shape=(y_shape_cv2[0], 1))
        self.pred_cv3 = np.zeros(shape=(y_shape_cv3[0], 1))
        self.mod = ""

    def Jchemo_lwplsr_fit(self):
        """Send data to Julia to fit lwplsr.

        Args:
            self.jl.x_train (DataFrame):
            self.jl.y_train (DataFrame):
            self.jl.x_test (DataFrame):
            self.jl.y_test (DataFrame):

        Returns:
            self.mod (Julia model): the prepared model
        """
        # launch Julia Jchemo lwplsr
        jl.seval("""
        using DataFrames
        using Pandas
        using Jchemo
        x_train |> Pandas.DataFrame |> DataFrames.DataFrame
        y_train |> Pandas.DataFrame |> DataFrames.DataFrame
        x_test |> Pandas.DataFrame |> DataFrames.DataFrame
        y_test |> Pandas.DataFrame |> DataFrames.DataFrame
        x_train_cv1 |> Pandas.DataFrame |> DataFrames.DataFrame
        y_train_cv1 |> Pandas.DataFrame |> DataFrames.DataFrame
        x_test_cv1 |> Pandas.DataFrame |> DataFrames.DataFrame
        y_test_cv1 |> Pandas.DataFrame |> DataFrames.DataFrame
        x_train_cv2 |> Pandas.DataFrame |> DataFrames.DataFrame
        y_train_cv2 |> Pandas.DataFrame |> DataFrames.DataFrame
        x_test_cv2 |> Pandas.DataFrame |> DataFrames.DataFrame
        y_test_cv2 |> Pandas.DataFrame |> DataFrames.DataFrame
        x_train_cv3 |> Pandas.DataFrame |> DataFrames.DataFrame
        y_train_cv3 |> Pandas.DataFrame |> DataFrames.DataFrame
        x_test_cv3 |> Pandas.DataFrame |> DataFrames.DataFrame
        y_test_cv3 |> Pandas.DataFrame |> DataFrames.DataFrame
        """)
        # Create LWPLSR model and fit
        jl.seval("""
        nlvdis = 5 ; metric = :mah
        h = 1 ; k = 200 ; nlv = 15 #; scal = true
        mod = Jchemo.model(Jchemo.lwplsr; nlvdis, metric, h, k, nlv)
        # Fit model
        Jchemo.fit!(mod, x_train, y_train)
        """)
        # CV model and fit
        jl.seval("""
        nlvdis = 5 ; metric = :mah
        h = 1 ; k = 200 ; nlv = 15 #; scal = true
        mod_cv1 = mod
        mod_cv2 = mod
        mod_cv3 = mod
        # Fit model
        Jchemo.fit!(mod_cv1, x_train_cv1, y_train_cv1)
        Jchemo.fit!(mod_cv2, x_train_cv2, y_train_cv2)
        Jchemo.fit!(mod_cv3, x_train_cv3, y_train_cv3)
        """)

        # jl.seval("""
        # mod = Jchemo.model(Jchemo.lwplsr)
        # res = Jchemo.gridcv(mod, x_train, y_train; segm, score = Jchemo.rmsep, pars, nlv, verbose = true).res
        # # u = findall(res.y1 .== minimum(res.y1))[1]
        # # mod = Jchemo.model(lwplsr; nlvdis = res.nlvdis[u], metric = res.metric[u], h = res.h[u], k = res.k[u], nlv = res.nlv[u]) ;
        # # Jchemo.fit!(mod, x_train, y_train)
        # """)
        self.mod = jl.mod

    def Jchemo_lwplsr_predict(self):
        """Send data to Julia to predict with lwplsr.

        Args:
            self.mod (Julia model): the prepared model
            self.jl.x_train (DataFrame):
            self.jl.y_train (DataFrame):
            self.jl.x_test (DataFrame):
            self.jl.y_test (DataFrame):

        Returns:
            self.pred_test (Julia DataFrame): predicted values on x_test
            self.pred_train (Julia DataFrame): predicted values on x_train
            self.pred_cv (Julia DataFrame): predicted values on x_train with Cross-Validation
        """
        # Predictions on x_test and store in self.pred
        self.pred_test = jl.seval("""
        println("start test predict")
        res = Jchemo.predict(mod, x_test)
        res.pred
        """)
        self.pred_train = jl.seval("""
        println("start train predict")
        res = Jchemo.predict(mod, x_train)
        res.pred
        """)
        self.pred_cv1 = jl.seval("""
        println("start test_cv1 predict")
        res = Jchemo.predict(mod_cv1, x_test_cv1)
        res.pred
        """)
        self.pred_cv2 = jl.seval("""
        println("start test_cv2 predict")
        res = Jchemo.predict(mod_cv2, x_test_cv2)
        res.pred
        """)
        self.pred_cv3 = jl.seval("""
        println("start test_cv3 predict")
        res = Jchemo.predict(mod_cv3, x_test_cv3)
        res.pred
        """)

    @property
    def pred_data_(self):
        # convert predicted data from x_test to Pandas DataFrame
        self.predicted_results_on_test = pd.DataFrame(self.pred_test)
        self.predicted_results_on_train = pd.DataFrame(self.pred_train)
        self.predicted_results_on_cv1 = pd.DataFrame(self.pred_cv1)
        self.predicted_results_on_cv2 = pd.DataFrame(self.pred_cv2)
        self.predicted_results_on_cv3 = pd.DataFrame(self.pred_cv3)
        return self.predicted_results_on_train, self.predicted_results_on_cv1, self.predicted_results_on_cv2, self.predicted_results_on_cv3, self.predicted_results_on_test

    @property
    def model_(self):
        return self.mod

    # @property
    # def metrics_(self):
    #     jl.pred_test = self.pred_test
    #     jl.seval("""
    #     using Jchemo
    #     """)
    #     scorermsep_test = jl.seval("""
    #         first(Jchemo.rmsep(pred_test, y_test))
    #         """)
    #     scoremr2_test = jl.seval("""
    #         first(Jchemo.r2(pred_test, y_test))
    #         """)
    #     scorerpd_test = jl.seval("""
    #         first(Jchemo.rpd(pred_test, y_test))
    #         """)
    #     scoremsep_test = jl.seval("""
    #         first(Jchemo.sep(pred_test, y_test))
    #         """)
    #     jl.pred_train = self.pred_train
    #     scorermsep_train = jl.seval("""
    #         first(Jchemo.rmsep(pred_train, y_train))
    #         """)
    #     scoremr2_train = jl.seval("""
    #         first(Jchemo.r2(pred_train, y_train))
    #         """)
    #     scorerpd_train = jl.seval("""
    #         first(Jchemo.rpd(pred_train, y_train))
    #         """)
    #     scoremsep_train = jl.seval("""
    #         first(Jchemo.sep(pred_train, y_train))
    #         """)
    #     jl.pred_cv = self.pred_cv
    #     scorermsep_cv = jl.seval("""
    #         first(Jchemo.rmsep(pred_cv, y_train))
    #         """)
    #     scoremr2_cv = jl.seval("""
    #         first(Jchemo.r2(pred_cv, y_train))
    #         """)
    #     scorerpd_cv = jl.seval("""
    #         first(Jchemo.rpd(pred_cv, y_train))
    #         """)
    #     scoremsep_cv = jl.seval("""
    #         first(Jchemo.sep(pred_cv, y_train))
    #         """)
    #
    #
    #     self.scores = pd.DataFrame([[scoremr2_test, scorermsep_test, scoremsep_test, scorerpd_test]], columns=['r2', 'rmsep', 'msep', 'rpd'], index=['test'])
    #     self.scores = pd.concat([self.scores, pd.DataFrame([[scoremr2_train, scorermsep_train, scoremsep_train, scorerpd_train]], columns=['r2', 'rmsep', 'msep', 'rpd'], index = ["train"]), pd.DataFrame([[scoremr2_cv, scorermsep_cv, scoremsep_cv, scorerpd_cv]], columns=['r2', 'rmsep', 'msep', 'rpd'], index = ["cv"])])#
    #     return self.scores
